'use strict';

Object.defineProperty(exports, "__esModule", {
    value: true
});

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

/**
 * Router should match given message to defined controller.
 * @constructor
 * @param {function} [resolveFunc] used as parser for message to match with controllers.
 */
var Router = exports.Router = function () {
    function Router() {
        var resolveFunc = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : null;

        _classCallCheck(this, Router);

        this.resolveFunc = resolveFunc;
        this.routes = {};
    }

    /**
     * `on` can register ControllerFunc to specified route name.
     * @param {string} name routing name for given ControllerFunc
     * @param {function} controllerFunc exacutable function when specified route is matched
    */


    _createClass(Router, [{
        key: 'on',
        value: function on(name, controllerFunc) {
            this.routes = this.routes || {};
            this.routes[name] = controllerFunc;
        }
        /**
         * `listener` provides listener function which can satisfy Chrome EventListener Func.
         * @return {function} EventListenerFunction
         * @see https://developer.chrome.com/extensions/runtime#event-onMessage
         */

    }, {
        key: 'listener',
        value: function listener() {
            return this.listen.bind(this);
        }

        // +++++ PRIVATE +++++

        /**
         * `listen` is an entry point for each `onMessage` invocation.
         * @private
         * parameters below are defined by `chorme` EventListener interface.
         * @param {any} message
         * @param {any} sender
         * @param {function} sendResponse
         *
         * TODO: some chrome interface can receive more than 1 arguments,
         * TODO: e.g. https://developer.chrome.com/apps/notifications#event-onButtonClicked
         * TODO: to handle this varieties of EventListener interfaces,
         * TODO: it should handle `arguments` and see if the last arg is Func.
         */

    }, {
        key: 'listen',
        value: function listen(message, sender) {
            var _this = this;

            var sendResponse = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : function () {};

            try {
                // Find matched controller.
                var controllerFunc = this.match(message, sender);

                // Return `Not Found` if no controller matched.
                if (!controllerFunc) {
                    // Not Found.
                    return true;
                }

                // TODO: Logger should be global for this module.
                // this.logger.info(this.constructor.name, handlerFunc.name);

                // Let controller create `response`.
                // TODO: some chrome interface can receive more than 1 arguments,
                // TODO: e.g. https://developer.chrome.com/apps/notifications#event-onButtonClicked
                var response = controllerFunc.call({ message: message, sender: sender }, message, sender);

                if (typeof response == 'undefined') {
                    sendResponse(this._formatResponse({
                        status: 500, // TODO: should it be 500?
                        message: '`' + (controllerFunc.name || '(anonymous controller)') + '`: Response should be defined. ex) return true;'
                    }));
                    return true;
                }

                // `response` can be Promise, but if it's not, just send it as a respone.
                if (!(response instanceof Promise)) {
                    sendResponse(this._formatResponse(response));
                    return true;
                }

                // If `response` is a Promise object, handle that promise for client.
                response.then(function (res) {
                    sendResponse(_this._formatResponse(res));
                }).catch(function (res) {
                    sendResponse(_this._formatResponse(res));
                });
                return true;
            } catch (err) {

                // If there is any error while executing, return 500.
                sendResponse(this._formatResponse({
                    status: 500, message: 'Unhandled Background Error: ' + String(err)
                }));

                return true;
            }
        }

        /**
         * `match` resolves message to routing name, and return registered controller.
         * @private
         * @param {any} message
         * @param {any} sender
         * @return {function} Matched Controller
         */

    }, {
        key: 'match',
        value: function match(message, sender) {
            var _resolveRoute2 = this._resolveRoute(message, sender),
                name = _resolveRoute2.name;

            var controllerFunc = this.routes[name];
            if (!controllerFunc) controllerFunc = this.constructor._NotFoundController;
            return controllerFunc;
        }

        /**
         * @private
         * `_resolveRoute` resolve given message to routing name if matched.
         * It prefers "resolveFunc" if it is given in constructor.
         * @param {any} message
         * @param {any} sender
         * @return {string} Matched Routing Name
         */

    }, {
        key: '_resolveRoute',
        value: function _resolveRoute(message, sender) {
            if (typeof this.resolveFunc == 'function') {
                return this.resolveFunc(message, sender);
            }
            return this._defaultResolveFunc(message, sender);
        }

        /**
         * @private
         * `_defaultResolveFunc` is a default resolve func,
         * which simply resolves routing name with priorities bellow
         * 1. just use message itself, if the message is `string`
         * 2. just use message.act or message.action, if the message is object.
         * @param {any} message
         * @return {string} Routing Name
         */

    }, {
        key: '_defaultResolveFunc',
        value: function _defaultResolveFunc(message) {
            if (typeof message == 'string') return {
                name: message, params: {}
            };
            if (typeof message.act == 'string') return {
                name: message.act, params: {}
            };
            if (typeof message.action == 'string') return {
                name: message.action, params: {}
            };
            return {
                name: '__notfound', params: {}
            };
        }

        /**
         * `_formatResponse` regulates response format.
         * @private
         * @param {any} response
         */

    }, {
        key: '_formatResponse',
        value: function _formatResponse(response) {
            if (response && Number.isInteger(response.status)) {
                return response;
            }
            return { status: 200, data: response.data || response };
        }

        /**
         * `_NotFoundController` is a default controller for unmatched routing.
         * @private
         * @static
         */

    }], [{
        key: '_NotFoundController',
        value: function _NotFoundController(message) {
            return {
                status: 404,
                message: 'routing not found for "' + (message.action || message.act) + '"'
            };
        }
    }]);

    return Router;
}();